//Name: Justin Fiedler
//Project: HW2:DecisionTree
//Course: CS 540

import java.util.*;

public class DecisionTree
{	private static boolean debug;
	private ExampleSet set;
	private Node root;
	private double errorRate;
	private ExampleSet misclassified = new ExampleSet(debug);

	/* DecisionTree
	 * constructs a new DecisionTree from the given ExampeSet
	 * with debug true = on, false = off
	 */
	public DecisionTree(ExampleSet set, boolean d)
	{	this(set, "", d);
	}

	/* DecisionTree
	 * recursively constructs a new DecisionTree from the given ExampeSet
	 */
	private DecisionTree(ExampleSet set, String branchValue, boolean d)
	{	debug = d;
		this.set = set;
		if(debug)
		{	System.out.println("data set: " + set);
			System.out.println("entropy: " + set.getEntropy());
		}
		if(set != null)
		{	if(set.getEntropy() != 0 && set.hasActiveAttribute())
			{	if(debug) System.out.println("decison");
				AttributeType bestAttribute = set.chooseSplitAttribute();
				if(debug) System.out.println("split feature: " + bestAttribute.label + "\n");
				root = new Node(branchValue, bestAttribute.label, set.getMajorityValue());
				for(int i = 0; i < bestAttribute.values.length; i++)
				{	//System.out.println(bestAttribute.values[i]);
					root.addChild((new DecisionTree(set.getSubSet(bestAttribute.values[i]), bestAttribute.values[i], d)).getRoot());
				}
			}
			else
			{	root = new Node(branchValue, "leaf", set.getMajorityValue());
				if(debug) System.out.println("leaf: " + root.majorityValue + "\n");
			}
		}
	}

	/* DecisionTree
	 * constructs a new DecisionTree with root node 'root'
	 */
	public DecisionTree(Node root)
	{	this.set = null;
		this.root = root;
	}

	/* getRoot
	 * returns the root node of the DecisionTree
	 */
	public Node getRoot()
	{	return root;
	}

	/* returns a String representation of
	 * the original tree
	 * test error/misclassification
	 * (prunes the tree) the pruned tree
	 */
	public String toString()
	{	//get string rep of the the tree from the training data
		String s = "\n\nTRAIN\n\n";
		if(root.splitAttribute == "leaf")
		{	s += root.majorityValue + "\n";
		}
		else
		{	s += root.splitAttribute + "\n";
			for(int i = 0; i < root.children.size(); i++)
			{	s += treeStructureString(((Node)root.children.get(i)), 4);
			}
		}

		// test set information
		s += "\n\n\nTEST\n\nerror rate: " + errorRate +
			 "\nmisclassifications: " + misclassified;

		//prune the tree on test set
		prune(root);

		//get string rep of the the pruned tree
		s += "\n\n\nPRUNE\n\n";
		if(root.splitAttribute == "leaf")
		{	s += root.majorityValue + "\n";
		}
		else
		{	s += root.splitAttribute + "\n";
			for(int i = 0; i < root.children.size(); i++)
			{	s += treeStructureString(((Node)root.children.get(i)), 4);
			}
		}
		return s;
	}

	/* prune
	 * prunes the tree based on the correctness of each
	 * node compared to its children in accordance to the test
	 * data
	 */
	public void prune(Node root)
	{	int childrenCorrectness = 0;
		for(int i = 0; i < root.children.size(); i++)
		{	Node child = (Node)root.children.get(i);
			prune(child);
			childrenCorrectness += child.correctlyClassified;
		}

		if(root.correctlyClassified >= childrenCorrectness)
		{	root.clearChildren();
			root.splitAttribute = "leaf";
		}
		else
		{	root.correctlyClassified = childrenCorrectness;
		}
	}

	/* testClassification
	 * tests the correctness of each example in the test set
	 * as classified in the decision tree
	 */
	public void testClassification(ExampleSet testset)
	{	for(int i = 0; i < testset.size(); i++)
		{	ExampleType ex = testset.getExample(i);
			testExample(ex, root);
		}
		errorRate = (double)misclassified.size() / testset.size();
	}

	/* testExample
	 * at each node the example passes through while
	 * being classified: if the nodes majority value would
	 * correctly classify the example the correctness of
	 * the node is increased by 1.
	 */
	public void testExample(ExampleType ex, Node root)
	{	if(ex.hasValue(root.majorityValue))
		{	root.correctlyClassified++;
		}
		if(root.hasChildren())
		{	for(int i = 0; i < root.children.size(); i++)
			{	Node node = (Node)root.children.get(i);
				if(ex.hasValue(node.branchValue))
				{	testExample(ex, node);
				}
			}
		}
		else
		{	if(!ex.hasValue(root.majorityValue))
			{	misclassified.addExample(ex);
			}
		}
	}

	/* treeStructureString
	 * returns a String representation of the tree
	 */
	public String treeStructureString(Node root, int indent)
	{	String s = "";
		for(int i = 0; i < indent; i++) s += " ";
		if(root.splitAttribute.equals("leaf"))
		{	s += root.branchValue + ": "+ root.majorityValue + "\n";
		}
		else
		{	s += root.branchValue + ": " +root.splitAttribute + "\n";
			for(int i = 0; i < root.children.size(); i++)
			{	Node child = (Node)root.children.get(i);
				s += treeStructureString(child, indent + 4);
			}
		}
		return s;
	}

	/* Node
	 * a class to represent individual Nodes in the DecisionTree
	 */
	private class Node
	{	public String branchValue;
		public String splitAttribute;
		public String majorityValue;
		public int correctlyClassified;
		public ArrayList children;

		/* Node
		 * constructs a new Node that is on the branch 'branchValue', splits on
		 * 'splitAttribute' and whose data set has a majority value of
		 * majority value.
		 */
		public Node(String branchValue, String splitAttribute, String majorityValue)
		{	this.branchValue = branchValue;
			this.splitAttribute = splitAttribute;
			this.majorityValue = majorityValue;
			correctlyClassified = 0;
			children = new ArrayList();
		}

		/* hasChildren
		 * returns true if the Node has children or otherwise false
		 */
		public boolean hasChildren()
		{	return children.size() > 0;
		}

		/* addChild
		 * adds child 'child' to the Node
		 */
		public void addChild(Node child)
		{	children.add(child);
		}

		/* clearChildren
		 * removes all children connected to Node
		 */
		public void clearChildren()
		{	children = new ArrayList();
		}

		/* toString
		 * returns a String representation of the Node
		 */
		public String toString()
		{	String s = "Node: \n" + splitAttribute + "\n"
					 + majorityValue + "\n" + children;
			return s;
		}
	}
}
