//Name: Justin Fiedler
//Project: HW2:ExampleSet
//Course: CS 540

import java.util.*;

/* ExampleSet
 * a class to represent a collection of examples with common attributes
 */
public class ExampleSet
{	ArrayList attributes;
	ArrayList examples;
	boolean debug;
	private static int count = 0;

	/* ExampleSet
	 * constructs a new empty ExampleSet
	 */
	public ExampleSet(boolean debug)
	{	this.attributes = new ArrayList();
		this.examples = new ArrayList();
		this.debug = debug;
	}

	/* addAttribute
	 * adds attribute 'label' w/ possible values 'values' to the set
	 */
	public void addAttribute(String label, String[] values)
        {     boolean active = true;
  	      if(label.equals("output")) active = false;
	      attributes.add(new AttributeType(label, values, active));
	}

	/* addAttribute
	 * adds attribute 'attribute' to the set
	 */
	private void addAttribute(AttributeType attribute)
	{	attributes.add(attribute);
	}

	/* addAttribute
	 * adds a ArrayList of attributes to the set
	 */
	public void addAttributeList(ArrayList attributes)
	{	for(int i = 0; i < attributes.size(); i++)
		{	this.attributes.add(attributes.get(i));
		}
	}

	/* addExample
	 * adds example values 'values' to the set
	 */
	public void addExample(String[] example)
	{	examples.add(new ExampleType(++count, example));
	}

	/* addExample
	 * adds example 'example' to the set
	 */
	public void addExample(ExampleType example)
	{	examples.add(example);
	}

	/* getSubSet
	 * returns an set of all examples in 'this' with the value
	 * 'attributeValue'.  The attribute of which this value belongs
	 * in the new set will be set inactive (active == false)
	 */
	public ExampleSet getSubSet(String attributeValue)
	{	ExampleSet subSet = new ExampleSet(debug);
		//find each example w/ value 'attributeValue
		for(int i = 0; i < examples.size(); i++)
		{	ExampleType ex = new ExampleType((ExampleType)examples.get(i));
			//System.out.print("Ex" + i + " " + ex + " ");
			//System.out.println(ex.hasValue(attributeValue));
			if(ex.hasValue(attributeValue))
			{	subSet.addExample(ex);
			}
		}

		//find the attribute that the set was split on and set it to inactive
		for(int i = 0; i < attributes.size(); i++)
		{	AttributeType att = new AttributeType((AttributeType)attributes.get(i));
			//System.out.print("Att" + i + " " + att + " ");
			//System.out.println(att.hasValue(attributeValue));

			if(att.hasValue(attributeValue))
			{	att.setActive(false);
			}
			subSet.attributes.add(att);
		}

		return subSet;
	}

	/* getEntropy
	 * returns the entropy of the set
	 */
	public double getEntropy()
	{	double entropy = 0;
		AttributeType currAtt = getAttribute("output");
		int[] valueCount = new int[currAtt.values.length];
		for(int i = 0; i < examples.size(); i++)
		{	ExampleType ex = (ExampleType)examples.get(i);
			for(int j = 0; j < currAtt.values.length; j++)
			{	if(ex.hasValue(currAtt.values[j]))
				{	valueCount[j]++;
				}
			}
		}

		//calculate entropy = sum 1 to n (-P(vi) log2 P(vi))
		//System.out.print("entropy: ");
		for(int i = 0; i < valueCount.length && examples.size() > 0; i++)
		{	double pi = (double)valueCount[i] / examples.size();
			if(pi != 0)
			{	entropy -= pi * (Math.log(pi) / Math.log(2));
			}
			//System.out.print("-" + valueCount[i] + "/" + examples.size());
			//System.out.print("*log(" + valueCount[i] + "/" + examples.size() + ")");

		}
		//System.out.println(" = " + entropy);
		return entropy;
	}

	/* getEntropy
	 * returns the entropy of the set if it where split on
	 * 'splitAttribute'
	 */
	public double getEntropy(String splitAttribute)
	{	double entropy = 0;
		AttributeType att = getAttribute(splitAttribute);
		for(int i = 0; i < att.values.length; i++)
		{	ExampleSet subSet = getSubSet(att.values[i]);
			entropy += (double)subSet.size() / size() * subSet.getEntropy();
		}
		return entropy;
	}

	/* getInformationGain
	 * returns the IG of the set if it where split on 'splitAttribute'
	 */
	public double getInformationGain(String splitAttribute)
	{	return getEntropy() - getEntropy(splitAttribute);
	}

	/* chooseSplitAttribtue
	 * returns the Attribute that achieves the highest IG as a split
	 * if there are no active attributes on which to split null is returned
	 */
	public AttributeType chooseSplitAttribute()
	{	double[] infoGain = new double[attributes.size()];

		//get InfoGain of each active attribute
		for(int i = 0; i < attributes.size() - 1; i++)
		{	AttributeType att = (AttributeType)attributes.get(i);
			if(att.active)
			{	infoGain[i] = getInformationGain(att.label);
				if(debug) System.out.println("  " + att.label + ": " + infoGain[i]);
			}
			else
			{	infoGain[i] = -1;
			}
		}

		//find the attribute with the highest IG
		int majority = 0;
		for(int i = 1; i < infoGain.length; i++)
		{	majority = infoGain[i] > infoGain[majority] ? i : majority;
		}
		if(((AttributeType)attributes.get(majority)).active)
		{	return (AttributeType)attributes.get(majority);
		}

		//no active attributes
		return null;
	}

        /* hasActiveAttribute
         * returns true if there exists an attribute that the set has not yet split on
         * else returns false
         */
        public boolean hasActiveAttribute()
        {   boolean hasActive = false;
	    for(int i = 0; i < attributes.size() && !hasActive; i++)
	    {   hasActive = ((AttributeType)attributes.get(i)).active;
	    }
            return hasActive;
        } 

	/* getAttribute
	 * returns the AttributeType 'label' in the list of attributes of the set
	 */
	private AttributeType getAttribute(String label)
	{	for(int i = 0; i < attributes.size(); i++)
		{	AttributeType att = (AttributeType) attributes.get(i);
			if(att.label.equals(label))
			return att;
		}
		System.out.println("No Such Attribute");
		System.exit(0);
		return null;
	}

	/* getMajorityValue
	 * returns majority output value in the ExampleSet
	 */
	public String getMajorityValue()
	{	AttributeType att = getAttribute("output");
		int[] valueCount = new int[att.values.length];
		for(int i = 0; i < examples.size(); i++)
		{	ExampleType ex = (ExampleType)examples.get(i);
			for(int j = 0; j < att.values.length; j++)
			{	if(ex.values[attributes.indexOf(att)].equals(att.values[j]))
				{	valueCount[j]++;
				}
			}
		}

		int majority = 0;
		for(int i = 0; i < valueCount.length; i++)
		{	majority = valueCount[i] > valueCount[majority] ? i : majority;
		}
		return att.values[majority];
	}

	/* getSize
	 * returns the number of examples in the set
	 */
	public int size()
	{	return examples.size();
	}

	/* getAttribute
	 * returns the 'i'th example in the set
	 */
	public ExampleType getExample(int i)
	{	return (ExampleType)examples.get(i);
	}

	/* toStringD
	 * returns a full String represetnation of the set including
	 * all examples/attributes and there labels/values.
	 */
	public String toStringD()
	{	String s = "Attributes\n";
		for(int i = 0; i < attributes.size(); i++)
		{	s += attributes.get(i) + "\n";
		}
		s += "\nExamples\n";
		for(int i = 0; i < examples.size(); i++)
		{	s += examples.get(i) + "\n";
		}
		return s;
	}

	/* toStringD
	 * returns a String represetnation of the set
	 * that displays the example #'s of the examplss in the set
	 */
	public String toString()
	{	String s = "";
		for(int i = 0; i < examples.size(); i++)
		{	s += ((ExampleType)examples.get(i)).num;
			if(i < examples.size() - 1) s += ",";
		}
		return s;
	}


}
