//Name: Justin Fiedler
//Project: HW2
//Course: CS 540

import java.io.*;

public class HW2
{	DecisionTree dtree;

	public static void main(String args[])
	{	boolean debug = false;

		String nametext = args[0];
		String train    = args[1];
		String test     = args[2];

		if(args.length == 4  && args[3].equals("-d"))
		{    debug = true;
		}
		
		new HW2(nametext, train, test, debug);
	}

	public HW2(String nametext, String train, String test, boolean debug)
	{
		//construct an ExampleSet from the train set information
		ExampleSet trainset =  new ExampleSet(debug);
		trainset = readAttributeList(nametext, trainset);
		trainset = readExampleSet(train, trainset);

		//System.out.println(trainset.toStringD());

		//Construct a DecisionTree from the trainset
		DecisionTree dtree = new DecisionTree(trainset, debug);

		//construct an ExampleSet from the test set information
		ExampleSet testset =  new ExampleSet(debug);
		testset = readAttributeList(nametext, testset);
		testset = readExampleSet(test, testset);

		//test the error and misclassifications of the test set
		dtree.testClassification(testset);

		//print the original tree, test error/misclassification, prune the tree,
		//print the pruned tree
		System.out.println(dtree);
	}

    /* readExampleSet
	 * accepts the name of a file from which to read a list of examples
     * using MyFileReader as well as an ExampleSet in which to place them
     */
	public ExampleSet readExampleSet(String fileName, ExampleSet set)
	{	String label;
		String values[] = new String[10];
		try
		{	MyFileReader fr = new MyFileReader(fileName, ",");
			String line = fr.readLine();
			while(line != null)
			{	int i;
				for(i = 0; fr.hasMoreTokens(); i++)
				{	if(i >= values.length)
					{	values = resizeArray(values, 2*values.length);
					}
					values[i] = fr.nextToken();
				}

				values = resizeArray(values, i);
				set.addExample(values);

				line = fr.readLine();
			}
		}
		catch(FileNotFoundException e)
		{	System.out.println(fileName + " not found");
			System.exit(0);
		}
		catch(IOException e)
		{	System.out.println("Error reading from " + fileName);
			System.exit(0);
		}
		return set;
	}

	/* readAttributeList
	 * accepts the name of a file from which to read a list of attributes
     * using MyFileReader as well as an ExampleSet on which this
     * attributes apply
     */
	public ExampleSet readAttributeList(String fileName, ExampleSet set)
	{	String attribute;
		String values[] = new String[10];
		try
		{	MyFileReader fr = new MyFileReader(fileName, ":\t,");
			String line = fr.readLine();
			while(line != null)
			{	attribute = fr.nextToken();
				int i;
				for(i = 0; fr.hasMoreTokens(); i++)
				{	if(i >= values.length)
					{	values = resizeArray(values, 2 * values.length);
					}
					values[i] = fr.nextToken();
				}

				values = resizeArray(values, i);
				set.addAttribute(attribute, values);

				line = fr.readLine();
			}
		}
		catch(FileNotFoundException e)
		{	System.out.println(fileName + " not found");
			System.exit(0);
		}
		catch(IOException e)
		{	System.out.println("Error reading from " + fileName);
			System.exit(0);
		}
		return set;
	}

	/* resizeArray
	 * returns an String array of size 'newSize' containing the
	 * contents of 'arr'
	 */
	private String[] resizeArray(String[] arr, int newSize)
	{	String[] temp = new String[newSize];
		for(int i = 0; (i < arr.length && i < newSize); i++)
		{	temp[i] = arr[i];
		}
		return temp;
	}

}
